# ==============================================================================
# Medical Image Registration ToolKit (MIRTK)
#
# Copyright 2013-2016 Imperial College London
# Copyright 2013-2016 Andreas Schuh
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

##############################################################################
# @file  MIRTKConfig.cmake
# @brief CMake package configuration file.
#
# @note The MIRTKConfig.cmake file is automatically generated
#       by BASIS from the template file Config.cmake.in which is part of
#       the BASIS package.
#
# Variables in this file have to be prefixed by the name of this package.
# In the default package configuration file of BASIS, which can be included
# here using the \@BASIS_CONFIG\@ variable, this prefix is defined as @c NS
# variable. Moreover, the macro basis_set_config() should be used for
# convenience to set any configuration variable, e.g.,
#
# @code
# basis_set_config (INCLUDE_DIR "\@INCLUDE_DIR_CONFIG\@")
# @endcode
#
# where the variable @c INCLUDE_DIR_CONFIG is set in the ConfigSettings.cmake
# file of the BASIS installation. Note that this statement will set the
# variable ${NS}INCLUDE_DIR, where by default @c NS is the package name
# followed by an underscore (_).
#
# See the Config.cmake.in file of BASIS for details.
#
# @ingroup BasisConfig
##############################################################################

## @addtogroup BasisConfig
# @{

# ============================================================================
# List of requested modules
# ============================================================================

if (MIRTK_FIND_COMPONENTS)
  set(MIRTK_MODULES_REQUESTED "${MIRTK_FIND_COMPONENTS}")
  foreach (_CMP IN ITEMS BASIS CMake)
    list(FIND MIRTK_MODULES_REQUESTED ${_CMP} _IDX)
    if (_IDX EQUAL -1)
      set(_INCLUDE_${_CMP}_MODULES FALSE)
    else ()
      list(REMOVE_ITEM MIRTK_MODULES_REQUESTED ${_CMP})
      set(_INCLUDE_${_CMP}_MODULES TRUE)
    endif ()
  endforeach ()
  unset(_CMP)
  unset(_IDX)
endif ()


# ============================================================================
# namespace
# ============================================================================

# prefix used for variable names
set (NS "MIRTK_")

# allow caller to change namespace - used by projects with modules
if (${NS}CONFIG_PREFIX)
  set (NS "${${NS}CONFIG_PREFIX}_")
endif ()

# ----------------------------------------------------------------------------
macro (basis_set_config NAME)
  set (${NS}${NAME} ${ARGN})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_copy_config NAME SRC)
  set (${NS}${NAME} ${${NS}${SRC}})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_append_config NAME)
  list (APPEND ${NS}${NAME} ${ARGN})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_find_config NAME VALUE IDX)
  list (FIND ${NS}${NAME} "${VALUE}" ${IDX})
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_remove_duplicates_config NAME)
  if (${NS}${NAME})
    list (REMOVE_DUPLICATES ${NS}${NAME})
  endif ()
endmacro ()

# ----------------------------------------------------------------------------
macro (basis_remove_item_config NAME ITEM)
  if (${NS}${NAME})
    list (REMOVE_ITEM ${NS}${NAME} "${ITEM}")
  endif ()
endmacro ()

# ----------------------------------------------------------------------------
if (NOT COMMAND basis_is_cached)
  macro (basis_is_cached VAR ENTRY)
    if (DEFINED ${ENTRY})
      get_property (${VAR} CACHE ${ENTRY} PROPERTY TYPE SET)
    else ()
      set (${VAR} FALSE)
    endif ()
  endmacro ()
endif ()

# ----------------------------------------------------------------------------
if (NOT COMMAND basis_set_or_update_value)
  macro (basis_set_or_update_value VAR)
    basis_is_cached (_CACHED ${VAR})
    if (_CACHED)
      if (${ARGC} GREATER 1)
        set_property (CACHE ${VAR} PROPERTY VALUE ${ARGN})
      else ()
        set (${VAR} "" CACHE INTERNAL "" FORCE)
      endif ()
    else ()
      set (${VAR} ${ARGN})
    endif ()
    unset (_CACHED)
  endmacro ()
endif ()

# ============================================================================
# general
# ============================================================================

## @brief Installation prefix or root of build tree.
basis_set_config (INSTALL_PREFIX "${CMAKE_CURRENT_LIST_DIR}/../../../")
get_filename_component (${NS}INSTALL_PREFIX "${${NS}INSTALL_PREFIX}" ABSOLUTE)

## @brief Project use file to be included by other projects.
basis_set_config (USE_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKUse.cmake")

## @brief Directory of example files.
basis_set_config (EXAMPLE_DIR "")

# ============================================================================
# libraries
# ============================================================================

## @brief Directory of header files.
basis_set_config (INCLUDE_DIR "${${NS}INSTALL_PREFIX}/include")

## @brief Directory of header files including dependencies.
basis_set_config    (INCLUDE_DIRS) # to have a set recognized by Doxygen filter
basis_copy_config   (INCLUDE_DIRS INCLUDE_DIR)
basis_append_config (INCLUDE_DIRS "")

## @brief Directory of libraries.
basis_set_config (LIBRARY_DIR "${${NS}INSTALL_PREFIX}/lib/mirtk")

## @brief Directory of libraries including dependencies.
basis_set_config    (LIBRARY_DIRS) # to have a set() recognized by Doxygen filter
basis_copy_config   (LIBRARY_DIRS LIBRARY_DIR)
basis_append_config (LIBRARY_DIRS "")

## @brief Search path for Python modules.
basis_set_config (PYTHONPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/python")

## @brief Search path for Jython modules.
basis_set_config (JYTHONPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/jython")

## @brief Search path for Perl modules.
basis_set_config (PERL5LIB "${${NS}INSTALL_PREFIX}/lib/mirtk/perl")

## @brief Search path for MATLAB modules.
basis_set_config (MATLABPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/matlab")

## @brief Search path for Bash modules.
#
# See the import() function of the BASIS Bash utilities.
basis_set_config (BASHPATH "${${NS}INSTALL_PREFIX}/lib/mirtk/bash")

# ============================================================================
# exports
# ============================================================================

## @brief Absolute path of exports file to be included by other projects.
basis_set_config (EXPORTS_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKTargets.cmake")

## @brief Absolute path of custom exports file to be included by other projects.
basis_set_config (CUSTOM_EXPORTS_FILE "${CMAKE_CURRENT_LIST_DIR}/MIRTKCustomTargets.cmake")

# ============================================================================
# modules
# ============================================================================

## @brief Directory containing module configuration files
basis_set_config (MODULES_DIR "${${NS}INSTALL_PREFIX}/lib/cmake/mirtk")

## @brief List of known project modules.
basis_set_config (MODULES "Common;IO;Image;Numerics;PointSet;Registration;Transformation;Deformable;Mapping;DrawEM;Scripting;Viewer;Applications")

## @brief List of found requested modules.
basis_set_config (MODULES_FOUND)

## @brief List of not found optional modules.
basis_set_config (MODULES_NOTFOUND)

## @brief List of available project modules.
basis_set_config (MODULES_ENABLED)
file (GLOB _CONFIG_FILES "${${NS}MODULES_DIR}/MIRTK*Config.cmake")
foreach (_CONFIG_FILE IN LISTS _CONFIG_FILES)
  if (_CONFIG_FILE MATCHES "/MIRTK([^ /][^ /]*)Config.cmake")
    set (_MODULE_NAME ${CMAKE_MATCH_1})
    basis_append_config (MODULES         ${_MODULE_NAME})
    basis_append_config (MODULES_ENABLED ${_MODULE_NAME})
  endif ()
endforeach ()

basis_remove_duplicates_config (MODULES)
basis_remove_duplicates_config (MODULES_ENABLED)

unset (_CONFIG_FILES)
unset (_CONFIG_FILE)

## @brief List of modules requested by external package using the
#         @c COMPONENTS argument of
#         <a href="http://www.cmake.org/cmake/help/cmake-2-8-docs.html#command:find_package">
#         find_package()</a>.
if (NOT DEFINED ${NS}MODULES_REQUESTED)
  if (${NS}FIND_COMPONENTS)
    basis_copy_config (MODULES_REQUESTED FIND_COMPONENTS)
  endif ()
  if (NOT ${NS}MODULES_REQUESTED)
    basis_copy_config (MODULES_REQUESTED MODULES_ENABLED)
  endif ()
endif ()
basis_remove_duplicates_config (MODULES_REQUESTED)

set (_PACKAGE_NS "${NS}")
foreach (_MODULE_NAME IN LISTS ${NS}MODULES_REQUESTED)
  # check if requested module is available
  basis_find_config (MODULES_ENABLED ${_MODULE_NAME} _MODULE_IDX)
  if (_MODULE_IDX EQUAL -1)
    # set <Pkg>_<Module>_FOUND = FALSE
    basis_set_config (${_MODULE_NAME}_FOUND FALSE)
    # append to list of not found requested modules
    basis_append_config (MODULES_NOTFOUND ${_MODULE_NAME})
  else ()
    set (_MODULE_NS "${NS}${_MODULE_NAME}_")
    # include configuration file of module
    set (${_MODULE_NAME}_CONFIG_PREFIX "${_MODULE_NS}")
    include ("${${NS}MODULES_DIR}/MIRTK${_MODULE_NAME}Config.cmake")
    unset (${_MODULE_NAME}_CONFIG_PREFIX)
    set (NS "${_PACKAGE_NS}") # restore NS of this package config
    # append include directories
    if (${_MODULE_NS}INCLUDE_DIRS)
      basis_append_config (INCLUDE_DIRS ${${_MODULE_NS}INCLUDE_DIRS})
    elseif (${_MODULE_NS}INCLUDE_DIR)
      basis_append_config (INCLUDE_DIRS ${${_MODULE_NS}INCLUDE_DIR})
    endif ()
    # append library directories
    if (${_MODULE_NS}LIBRARY_DIRS)
      basis_append_config (LIBRARY_DIRS ${${_MODULE_NS}LIBRARY_DIRS})
    elseif (${_MODULE_NS}LIBARY_DIR)
      basis_append_config (LIBRARY_DIRS ${${_MODULE_NS}LIBRARY_DIR})
    endif ()
    # set <Pkg>_<Module>_FOUND = TRUE
    basis_set_config (${_MODULE_NAME}_FOUND TRUE)
    # append to list of found requested modules
    basis_append_config (MODULES_FOUND ${_MODULE_NAME})
  endif ()
endforeach ()
unset (_PACKAGE_NS)

if (${NS}MODULES_NOTFOUND)
  set (_REQUIRED)
  set (_OPTIONAL)
  set (_MISSING)
  foreach (_MODULE_NAME IN LISTS ${NS}MODULES_REQUESTED)
    if (${NS}FIND_REQUIRED_${_MODULE_NAME})
      list (APPEND _REQUIRED ${_MODULE_NAME})
      if (NOT ${NS}${_MODULE_NAME}_FOUND)
        list (APPEND _MISSING ${_MODULE_NAME})
      endif ()
    else ()
      list (APPEND _OPTIONAL ${_MODULE_NAME})
    endif ()
  endforeach ()
  if (_MISSING)
    message(FATAL_ERROR "Could not find all required MIRTK components!"
                        " Rebuild and/or install the MIRTK package with"
                        " the following modules enabled by setting MODULE_<name>"
                        " to ON when configuring the build of the package with CMake."
                        "\nOptional modules = [${_OPTIONAL}]"
                        "\nRequired modules = [${_REQUIRED}]"
                        "\nMissing  modules = [${_MISSING}]\n")
  endif ()
  unset (_REQUIRED)
  unset (_OPTIONAL)
  unset (_MISSING)
endif ()

unset (_MODULE_NAME)
unset (_MODULE_IDX)
unset (_MODULE_NS)

# ============================================================================
# clean INCLUDE_DIRS and LIBRARY_DIRS
# ============================================================================

basis_remove_duplicates_config (INCLUDE_DIR)
basis_remove_duplicates_config (INCLUDE_DIRS)
basis_remove_duplicates_config (LIBRARY_DIR)
basis_remove_duplicates_config (LIBRARY_DIRS)

basis_remove_item_config (INCLUDE_DIRS "")
basis_remove_item_config (LIBRARY_DIRS "")

# ============================================================================
# dependencies
# ============================================================================



# ============================================================================
# CMake Modules
# ============================================================================

## @brief Directory containing the CMake BASIS Modules
set(BASIS_MODULE_PATH "${${NS}INSTALL_PREFIX}/share/mirtk/cmake")

if (EXISTS "${BASIS_MODULE_PATH}/BasisTools.cmake")
  set(MIRTK_BASIS_FOUND TRUE)
else ()
  set(MIRTK_BASIS_FOUND FALSE)
endif ()
if (NOT MIRTK_BASIS_FOUND AND (MIRTK_FIND_REQUIRED_BASIS OR MIRTK_FIND_REQUIRED_CMake))
  message (FATAL_ERROR "Required CMake BASIS Modules of package MIRTK"
                       " could not be found in ${BASIS_MODULE_PATH}!")
endif ()

## @brief Directory containing the MIRTK CMake Modules
set(MIRTK_MODULE_PATH "${${NS}INSTALL_PREFIX}/share/mirtk/cmake")

if (EXISTS "${MIRTK_MODULE_PATH}/mirtkTools.cmake")
  set(MIRTK_CMake_FOUND TRUE)
else ()
  set(MIRTK_CMake_FOUND FALSE)
endif ()
if (NOT MIRTK_CMake_FOUND AND MIRTK_FIND_REQUIRED_CMake)
  message (FATAL_ERROR "Required CMake Modules of package MIRTK"
                       " could not be found in ${MIRTK_MODULE_PATH}!")
endif ()
if (NOT MIRTK_BASIS_FOUND)
  set(MIRTK_CMake_FOUND FALSE)
endif ()

# include requested CMake modules
if (_INCLUDE_BASIS_MODULES OR _INCLUDE_CMake_MODULES)
  if (MIRTK_CMake_FOUND)
    list(APPEND CMAKE_MODULE_PATH "${MIRTK_MODULE_PATH}")
  endif ()
  if (MIRTK_BASIS_FOUND)
    list(APPEND CMAKE_MODULE_PATH "${BASIS_MODULE_PATH}")
  endif ()
  if (_INCLUDE_CMake_MODULES AND MIRTK_CMake_FOUND)
    include(mirtkPolicies NO_POLICY_SCOPE)
  endif ()
  if (MIRTK_BASIS_FOUND)
    include(BasisSettings)
    include(BasisTools)
  endif ()
  if (_INCLUDE_CMake_MODULES AND MIRTK_CMake_FOUND)
    include(mirtkTools)
  endif ()
endif ()

unset(_INCLUDE_BASIS_MODULES)
unset(_INCLUDE_CMake_MODULES)

# ============================================================================
# Common settings
# ============================================================================

## @brief Directory containing the MIRTK command executables
set(MIRTK_TOOLS_DIR "${${NS}INSTALL_PREFIX}/")

## @brief Whether to build shared libraries by default
if (NOT DEFINED BUILD_SHARED_LIBS_DEFAULT)
  set(BUILD_SHARED_LIBS_DEFAULT ON)
endif ()

## @brief Whether to enable profiling of code execution by default
if (NOT DEFINED WITH_PROFILING_DEFAULT)
  set(WITH_PROFILING_DEFAULT ON)
endif ()

## @brief Whether to build applications by default
if (NOT DEFINED BUILD_APPLICATIONS_DEFAULT)
  set(BUILD_APPLICATIONS_DEFAULT ON)
endif ()

## @brief Default for WITH_TBB
if (NOT DEFINED WITH_TBB_DEFAULT)
  set(WITH_TBB_DEFAULT true)
endif ()

## @brief Default for WITH_VTK
if (NOT DEFINED WITH_VTK_DEFAULT)
  set(WITH_VTK_DEFAULT true)
endif ()

## @brief Default for WITH_ZLIB
if (NOT DEFINED WITH_ZLIB_DEFAULT)
  set(WITH_ZLIB_DEFAULT ON)
endif ()


## @}
# end of Doxygen group - keep at end of file
